/*  Sony RM-X2S plugin for xmms by Dirk Jagdmann <doj@cubic.org>
 *                                 http://llg.cubic.org
 *
 *  XMMS is Copyright (C) 1998-2000 Peter Alm, Mikael Alm, Olle
 *  Hallnas, Thomas Nilsson and 4Front Technologies
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.  */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "sony.h"

/** main window widget */
static GtkWidget *mainwin=0;
/** popup window widget */
static GtkWidget *popupwin=0;
/** status bar widget */
static GtkWidget *status=0;
/** array holding the button widgets of the main window */
static GtkWidget *joybutton[SONY_NUM];
/** strings for the joybuttons above */
static gchar *joybutton_str[SONY_NUM] =
{
  "Eject",
  "Play/Pause",
  "Stop",
  "Next Song",
  "Previos Song",
  "Volume Up",
  "Volume Down",
  "Repeat",
  "Shuffle",
  "Always on Top",
  "Show Equalizer Window",
  "Show Main Window",
  "Show Playlist Window",
  "Show Plugin Preferences"
};

/** a local copy of the configuration */
static sonyrm_config local_cfg;	
/** flag to remember state of the program */
static gboolean wasactive=FALSE; 

/** This flag tells the thread to keep running. If set to false, the
   thread will exit */
static volatile gboolean sonyrm_joy_keep_going = FALSE;

/** The thread to read configuration values */
static pthread_t sonyrm_joy_thread;

/** update status bar 
    @param text string to display
*/
void sonyrm_addstatus(const gchar *text)
{
  if(status)
    gtk_statusbar_push(GTK_STATUSBAR(status),
		       gtk_statusbar_get_context_id(GTK_STATUSBAR(status), text),
		       text);
}

/** callback function to update the local configuration with an edited device entry 
    @param w ignored
    @param data string containing the new device name
*/
static void sonyrm_device_changed(GtkWidget * w, gpointer data)
{
  g_free(local_cfg.device);
  local_cfg.device = g_strdup(gtk_entry_get_text(GTK_ENTRY(data)));
  sonyrm_addstatus(_("Now calibrate the joystick"));
}

/** The button of the input dialog was pressed. The thread is
   terminated and the popup window destroyed
   @param w ignored
   @param data event number as defined by enum sonyrm_cmd_en (note, no pointer, we pass the value directly here!)
*/
static void sonyrm_joy_cb(GtkWidget * w, gpointer data)
{
  int eventnum=(int)data;
  gchar *str;

  sonyrm_joy_keep_going = FALSE;
  pthread_join(sonyrm_joy_thread, NULL);

  /* doj hier den text setzen */
  str=g_strdup_printf("%s: %i", joybutton_str[eventnum], local_cfg.command[eventnum]);
  gtk_widget_set(joybutton[eventnum], "label", str);

  /*sonyrm_addstatus((gchar *)data);*/
  if(popupwin)
    gtk_widget_destroy(popupwin);

#ifdef debug
  sonyrm_dumpconfig(&local_cfg);
#endif
}

/** read an event and store it in arg 
    @param eventnum event number as defined by enum sonyrm_cmd_en (note, no pointer, we pass the value directly here!)
*/
static void *sonyrm_joy_thread_routine(void *eventnum)
{
  int fd=-1, value=SONY_NIL, s=-1;
  struct js_event js;
  fd_set set;
  struct timeval tv;
  int initialvalue=SONY_NIL;
  gboolean flag=TRUE;

  if((fd=sonyrm_open(&local_cfg)) < 0)
    return eventnum;

  /* and we have to make sure the thread can be killed from outside. */
  /* this is managed by the global sonyrm_joy_keep_going flag */
  while(sonyrm_joy_keep_going && flag)
    {
      FD_ZERO(&set);
      FD_SET(fd, &set);

      tv.tv_sec = 0;
      tv.tv_usec = 10000;

      s=select(fd+1, &set, NULL, NULL, &tv);
      if(s<0)
	{
	  perror("error while selecting");
	  sonyrm_joy_keep_going = FALSE;
	  flag=FALSE;	  
	  continue;
	}

      if (FD_ISSET(fd, &set))
	{
	  int newvalue=SONY_NIL;

	  if(read(fd, &js, sizeof(struct js_event)) != sizeof(struct js_event))
	    {
	      perror("Sonystick Control: could not read");
	      sonyrm_joy_keep_going = FALSE;
	      flag=FALSE;	  
	      continue;
	    }

	  newvalue=js.value&local_cfg.bitmask; /* mask noise bits */
#ifdef debug
	  printf("sonyrm_configure(): type: 0x%02X value: %i mask: %#x masked: %i initialvalue: %i\n", js.type, js.value, local_cfg.bitmask, newvalue, initialvalue);
#endif

	  if(js.type==(JS_EVENT_AXIS|JS_EVENT_INIT))
	    {
	      initialvalue=newvalue; /* store initial value of "home" position */
#ifdef debug
	      printf("sonyrm_configure(): JS_EVENT_AXIS|JS_EVENT_INIT\n");
#endif
	    }

	  if(js.type==(JS_EVENT_BUTTON|JS_EVENT_INIT)) /* ignore initial button position */
	    {
#ifdef debug
	      printf("sonyrm_configure(): JS_EVENT_BUTTON|JS_EVENT_INIT\n");
#endif
	    }

	  if(js.type==JS_EVENT_BUTTON)
	    {
	      value=SONY_BUTTON; /* the button gets a special value */
	      flag=FALSE;
#ifdef debug
	      printf("sonyrm_configure(): JS_EVENT_BUTTON\n");
#endif
	    }

	  if(js.type==JS_EVENT_AXIS && newvalue!=initialvalue) /* an event occured */
	    {
	      value=newvalue;
	      flag=FALSE;
#ifdef debug
	      printf("sonyrm_configure(): JS_EVENT_AXIS\n");
#endif
	    }

	}
    }

  if(!flag)			/* when the upper loop has found an event */
    {
      putchar(7); fflush(stdout); /* beep */

      /* read any more events until this thread is killed by the popup window */
      sonyrm_joy_keep_going=TRUE;
      while(sonyrm_joy_keep_going)
	{
	  FD_ZERO(&set);
	  FD_SET(fd, &set);
	  tv.tv_sec = 0;
	  tv.tv_usec = 10000;
	  s=select(fd+1, &set, NULL, NULL, &tv);
	  if(s<0)
	    {
	      perror("error while selecting");
	      sonyrm_joy_keep_going = FALSE;
	      continue;
	    }
	  if (FD_ISSET(fd, &set))
	    {
	      if(read(fd, &js, sizeof(struct js_event)) != sizeof(struct js_event))
		{
		  perror("Sonystick Control: could not read dangling");
		  sonyrm_joy_keep_going = FALSE;
		  continue;
		}
#ifdef debug
	      printf("dangling: %i\n", js.value);
#endif	 
	    }
	}
    }

  if(close(fd) < 0)
    perror("Sonystick Control: could not close");

#ifdef debug
  printf("\nsonyrm_configure(): value: %i\n", value);
#endif

  if(value!=SONY_NIL)
    local_cfg.command[(int)eventnum]=value; /* we got a new value */

  return eventnum;
}

/** create the popup window and create the thread to record a new event 
    @param w ignored
    @param eventnum event number as defined by enum sonyrm_cmd_en (note, no pointer, we pass the value directly here!)
*/
static void sonyrm_joy_prepare(GtkWidget * w, gpointer eventnum)
{
  GtkWidget *mainbox, *label, *button;

  popupwin=gtk_window_new(GTK_WINDOW_DIALOG);
  gtk_window_set_title(GTK_WINDOW(popupwin), _("Press the button"));
  gtk_window_set_position(GTK_WINDOW(popupwin), GTK_WIN_POS_MOUSE);
  gtk_window_set_policy(GTK_WINDOW(popupwin), FALSE, FALSE, FALSE);
  gtk_signal_connect(GTK_OBJECT(popupwin), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &popupwin);
  gtk_container_border_width(GTK_CONTAINER(popupwin), 10);
  gtk_window_set_default_size(GTK_WINDOW(popupwin), 300, 30);
  gtk_widget_show(popupwin);

  mainbox = gtk_vbox_new(FALSE, 10);
  gtk_container_add(GTK_CONTAINER(popupwin), mainbox);
  gtk_widget_show(mainbox);

  label=gtk_label_new(_("Press the appropriate button on your Sonystick"));
  gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), label);
  gtk_widget_show(label);

  button = gtk_button_new_with_label(_("OK"));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_cb), eventnum);
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), button);
  gtk_widget_grab_default(button);
  gtk_widget_show(button);

  button = gtk_button_new_with_label(_("Cancel"));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_cb), eventnum);
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), button);
  gtk_widget_show(button);

  sonyrm_addstatus(_("press the button on your Sonystick"));

  sonyrm_joy_keep_going = TRUE;
  if(pthread_create(&sonyrm_joy_thread, NULL, sonyrm_joy_thread_routine, eventnum) < 0)
    perror("could not create thread");
}

/** clears the local configuration structure
    @param w ignored
    @param data ignored
*/
static void sonyrm_joy_clear(GtkWidget * w, gpointer data)
{
  int i;
  for(i=0; i<SONY_NUM; i++)
    local_cfg.command[i]=SONY_NIL;
  sonyrm_addstatus(_("Now configure the buttons again"));
}

/** updates the bitmask field in the configuration structure according to the precision radio buttons.
    @param w ignored
    @param data a string containing the radio button number
*/
static void sonyrm_precision(GtkWidget * w, gpointer data)
{
  local_cfg.bitmask=SONY_INITIAL_BITMASK << atoi((const char *)data);
  sonyrm_addstatus(_("Now configure the buttons"));
#ifdef debug
  printf("bitmask: %s -> %#x\n", (const char *)data, local_cfg.bitmask);
#endif
}

/** The OK button was pressed. The new configuration is updated and saved and the sonyrm main thread is resumed if neccessary.
    @param w ignored
    @param data ignored
 */
static void sonyrm_ok_cb(GtkWidget * w, gpointer data)
{
  sonyrm_copy(&sonyrm_cfg, &local_cfg);	/* copy local configuration to glocal */
  sonyrm_save_config();
  gtk_widget_destroy(mainwin);

  if(wasactive)
    sonyrm_init();		/* activate the plugin again */
}

/** The CANCEL button was pressed. The sonyrm main thread is resumed if neccessary.
    @param w ignored
    @param data ignored
*/
static void sonyrm_cancel_cb(GtkWidget * w, gpointer data)
{
  gtk_widget_destroy(mainwin);

  if(wasactive)
    sonyrm_init();
}

/** checks if a device exists
    @param device string containing device name
    @return true if device exists
*/
static gboolean checkdevice(const gchar *device)
{
  struct stat buf;
  if(stat(device, &buf) < 0)
    {
#ifdef debug
      perror("Sonystick Control: could not stat");
#endif
      return FALSE;
    }

  if(S_ISCHR(buf.st_mode))
    return TRUE;

  return FALSE;
}

/** sets up the configuration window. The sonyrm main thread is stopped. */
void sonyrm_configure()
{
  GtkWidget *mainbox, *box, *label, *button, *table, *e_device;
  guint16 mask=SONY_INITIAL_BITMASK; /* needed for the precision radio buttons */

  sonyrm_read_config();	/* read configuration from disk */
  sonyrm_copy(&local_cfg, &sonyrm_cfg); /* and make a local copy */

  if((wasactive=sonyrm_active())) /* store state of plugin */
    sonyrm_cleanup();		/* if it was running, stop it */

  /* check or guess the device */
  if(!checkdevice(local_cfg.device))
    {
      if(checkdevice("/dev/js0"))
	{
	  local_cfg.device="/dev/js0";
	}
      else if(checkdevice("/dev/input/js0"))
	{
	  local_cfg.device="/dev/input/js0";
	}
    }

  /* create main window */
  mainwin = gtk_window_new(GTK_WINDOW_DIALOG);
  gtk_signal_connect(GTK_OBJECT(mainwin), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &mainwin);
  gtk_window_set_title(GTK_WINDOW(mainwin), _("XMMS Sonystick Configuration"));
  gtk_window_set_policy(GTK_WINDOW(mainwin), FALSE, FALSE, TRUE);
  gtk_window_set_position(GTK_WINDOW(mainwin), GTK_WIN_POS_MOUSE);
  gtk_container_border_width(GTK_CONTAINER(mainwin), 10);
  gtk_widget_show(mainwin);

  /* vertical box to hold everything */
  mainbox = gtk_vbox_new(FALSE, 10);
  gtk_container_add(GTK_CONTAINER(mainwin), mainbox);
  gtk_widget_show(mainbox);

  /* device */
  box=gtk_hbox_new(FALSE, 10);
  gtk_container_add(GTK_CONTAINER(mainbox), box);
  gtk_widget_show(box);

  label=gtk_label_new(_("Device:"));
  gtk_box_pack_start_defaults(GTK_BOX(box), label);
  gtk_widget_show(label);

  e_device = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(e_device), local_cfg.device);
  gtk_signal_connect(GTK_OBJECT(e_device), "changed", GTK_SIGNAL_FUNC(sonyrm_device_changed), e_device);
  gtk_box_pack_start_defaults(GTK_BOX(box), e_device);
  gtk_widget_show(e_device);

  /* Calibration Button */
  button=gtk_button_new_with_label(_("Calibrate Sonystick"));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_calibrate_init), &local_cfg);
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), button);
  gtk_widget_show(button);

  /* Precision */
  box=gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), box);
  gtk_widget_show(box);

  label=gtk_label_new(_("Precision:"));
  gtk_box_pack_start_defaults(GTK_BOX(box), label);
  gtk_widget_show(label);

  button = gtk_radio_button_new_with_label (NULL, "0");
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_precision), "0");
  gtk_box_pack_start_defaults(GTK_BOX(box), button);
  gtk_widget_show (button);
  if(local_cfg.bitmask == mask)
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
  mask<<=1;

  button = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON(button), "1");
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_precision), "1");
  gtk_box_pack_start_defaults(GTK_BOX(box), button);
  gtk_widget_show (button);
  if(local_cfg.bitmask == mask)
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
  mask<<=1;

  button = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON(button), "2");
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_precision), "2");
  gtk_box_pack_start_defaults(GTK_BOX(box), button);
  gtk_widget_show (button);
  if(local_cfg.bitmask == mask)
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
  mask<<=1;

  button = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON(button), "3");
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_precision), "3");
  gtk_box_pack_start_defaults(GTK_BOX(box), button);
  gtk_widget_show (button);
  if(local_cfg.bitmask == mask)
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
  mask<<=1;

  button = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON(button), "4");
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_precision), "4");
  gtk_box_pack_start_defaults(GTK_BOX(box), button);
  gtk_widget_show (button);
  if(local_cfg.bitmask == mask)
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
  mask<<=1;

  button = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON(button), "5");
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_precision), "5");
  gtk_box_pack_start_defaults(GTK_BOX(box), button);
  gtk_widget_show (button);
  if(local_cfg.bitmask == mask)
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
  mask<<=1;

  button = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON(button), "6");
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_precision), "6");
  gtk_box_pack_start_defaults(GTK_BOX(box), button);
  gtk_widget_show (button);
  if(local_cfg.bitmask == mask)
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
  mask<<=1;

  button = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON(button), "7");
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_precision), "7");
  gtk_box_pack_start_defaults(GTK_BOX(box), button);
  gtk_widget_show (button);
  if(local_cfg.bitmask == mask)
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
  mask<<=1;

  /* Buttons */
  table=gtk_table_new(8, 2, TRUE);
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), table);
  gtk_widget_show(table);

  button=joybutton[SONY_PLAY]=gtk_button_new_with_label(_(joybutton_str[SONY_PLAY]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_PLAY);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 0, 1, 0, 1);
  gtk_widget_show(button);

  button=joybutton[SONY_STOP]=gtk_button_new_with_label(_(joybutton_str[SONY_STOP]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_STOP);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 1, 2, 0, 1);
  gtk_widget_show(button);

  button=joybutton[SONY_NEXT]=gtk_button_new_with_label(_(joybutton_str[SONY_NEXT]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_NEXT);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 0, 1, 1, 2);
  gtk_widget_show(button);

  button=joybutton[SONY_PREV]=gtk_button_new_with_label(_(joybutton_str[SONY_PREV]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_PREV);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 1, 2, 1, 2);
  gtk_widget_show(button);

  button=joybutton[SONY_VOLUP]=gtk_button_new_with_label(_(joybutton_str[SONY_VOLUP]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_VOLUP);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 0, 1, 2, 3);
  gtk_widget_show(button);

  button=joybutton[SONY_VOLDOWN]=gtk_button_new_with_label(_(joybutton_str[SONY_VOLDOWN]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_VOLDOWN);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 1, 2, 2, 3);
  gtk_widget_show(button);

  button=joybutton[SONY_REPEAT]=gtk_button_new_with_label(_(joybutton_str[SONY_REPEAT]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_REPEAT);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 0, 1, 3, 4);
  gtk_widget_show(button);

  button=joybutton[SONY_SHUFFLE]=gtk_button_new_with_label(_(joybutton_str[SONY_SHUFFLE]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_SHUFFLE);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 1, 2, 3, 4);
  gtk_widget_show(button);

  button=joybutton[SONY_EJECT]=gtk_button_new_with_label(_(joybutton_str[SONY_EJECT]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_EJECT);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 0, 1, 4, 5);
  gtk_widget_show(button);

  button=joybutton[SONY_AOT]=gtk_button_new_with_label(_(joybutton_str[SONY_AOT]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_AOT);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 1, 2, 4, 5);
  gtk_widget_show(button);

  button=joybutton[SONY_EQWIN]=gtk_button_new_with_label(_(joybutton_str[SONY_EQWIN]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_EQWIN);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 0, 1, 5, 6);
  gtk_widget_show(button);

  button=joybutton[SONY_MAINWIN]=gtk_button_new_with_label(_(joybutton_str[SONY_MAINWIN]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_MAINWIN);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 1, 2, 5, 6);
  gtk_widget_show(button);

  button=joybutton[SONY_PLWIN]=gtk_button_new_with_label(_(joybutton_str[SONY_PLWIN]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_PLWIN);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 0, 1, 6, 7);
  gtk_widget_show(button);

  button=joybutton[SONY_PREFWIN]=gtk_button_new_with_label(_(joybutton_str[SONY_PREFWIN]));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_prepare), (void*)SONY_PREFWIN);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 1, 2, 6, 7);
  gtk_widget_show(button);

  button=gtk_button_new_with_label(_("Clear Everything"));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_joy_clear), NULL);
  gtk_table_attach_defaults(GTK_TABLE(table), button, 0, 2, 7, 8);
  gtk_widget_show(button);

  /* OK, Cancel */
  box = gtk_hbutton_box_new();
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), box);
  gtk_widget_show(box);

  button = gtk_button_new_with_label(_("Ok"));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_ok_cb), NULL);
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(box), button, TRUE, TRUE, 0);
  gtk_widget_grab_default(button);
  gtk_widget_show(button);

  button = gtk_button_new_with_label(_("Cancel"));
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_cancel_cb), NULL);
  gtk_box_pack_start(GTK_BOX(box), button, TRUE, TRUE, 0);
  gtk_widget_show(button);

  /* Statusbar */
  status=gtk_statusbar_new();
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), status);
  gtk_widget_show(status);
  sonyrm_addstatus(_("Check device settings"));
}
