/*  Sony RM-X2S plugin for xmms by Dirk Jagdmann <doj@cubic.org>
 *                                 http://llg.cubic.org
 *
 *  XMMS is Copyright (C) 1998-2000 Peter Alm, Mikael Alm, Olle
 *  Hallnas, Thomas Nilsson and 4Front Technologies
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.  */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "sony.h"

/** The thread handle */
static pthread_t sonyrm_thread;

/** This flag tells the thread to keep running. If set to false, the
   thread will exit */
static volatile gboolean keep_going = FALSE;

/* forward declaration */
static void *sonyrm_routine(void *arg);

/** XMMS Plugin structure */
GeneralPlugin sonyrm_gp =
{
  NULL,			/* handle */
  NULL,			/* filename */
  -1,			/* xmms_session */
  "Sony RM-X2S Control",
  sonyrm_init,
  sonyrm_about,
  sonyrm_configure,
  sonyrm_cleanup
};

/** initial Sonyrm config structure */
sonyrm_config sonyrm_cfg =
{
  "/dev/js0",			/* device */
  { 
    SONY_NIL,			/* SONY_EJECT */
    SONY_NIL,			/* SONY_PLAY */
    SONY_NIL,			/* SONY_STOP */
    SONY_NIL,			/* SONY_NEXT */
    SONY_NIL,			/* SONY_PREV */
    SONY_NIL,			/* SONY_VOLUP */
    SONY_NIL,			/* SONY_VOLDOWN */
    SONY_NIL,			/* SONY_REPEAT */
    SONY_NIL,			/* SONY_SHUFFLE */
    SONY_NIL,			/* SONY_AOT */
    SONY_NIL,			/* SONY_EQWIN */
    SONY_NIL,			/* SONY_MAINWIN */
    SONY_NIL,			/* SONY_PLWIN */
    SONY_NIL			/* SONY_PREFWIN */
  },
  SONY_INITIAL_BITMASK,		/* bitmask */
  {				/* struct js_corr */
    { 0,0,0,0,0,0,0,0 },
    0,
    0, /*SONY_IGNORE_CALIBRATION*/
  }
};

char *sonyrm_configstr[] =
{
  "SONY_EJECT", 
  "SONY_PLAY", 
  "SONY_STOP", 
  "SONY_NEXT", 
  "SONY_PREV", 
  "SONY_VOLUP", 
  "SONY_VOLDOWN",
  "SONY_REPEAT", 
  "SONY_SHUFFLE",
  "SONY_AOT", 
  "SONY_EQWIN", 
  "SONY_MAINWIN",
  "SONY_PLWIN", 
  "SONY_PREFWIN",
  "coef0",
  "coef1",
  "coef2",
  "coef3",
  "coef4",
  "coef5",
  "coef6",
  "coef7"
};


/** @return the XMMS plugin structure */
GeneralPlugin *get_gplugin_info(void)
{
  return &sonyrm_gp;
}

/** @return if the thread is running */
const gboolean sonyrm_active()
{
  return keep_going;
}

/** reads the configuration from the XMMS config file in the global config structure */
void sonyrm_read_config(void)
{
  ConfigFile *cfile = xmms_cfg_open_default_file();
  if (cfile)
    {
      int i;
      gint d;

      xmms_cfg_read_string(cfile, "sonystick", "device", &sonyrm_cfg.device);

      for(i=0; i<SONY_NUM; i++)
	xmms_cfg_read_int(cfile, "sonystick", sonyrm_configstr[i], &sonyrm_cfg.command[i]);

      /* take some care for the bitmask */
      xmms_cfg_read_int(cfile, "sonystick", "bitmask", &d); sonyrm_cfg.bitmask=d;
      sonyrm_cfg.bitmask&=SONY_INITIAL_BITMASK;

      xmms_cfg_read_int(cfile, "sonystick", "type", &d); sonyrm_cfg.correction.type=d;
      xmms_cfg_read_int(cfile, "sonystick", "prec", &d); sonyrm_cfg.correction.prec=d;
      for(i=0; i<8; i++)
	{
	  xmms_cfg_read_int(cfile, "sonystick", sonyrm_configstr[SONY_NUM+i], &d); 
	  sonyrm_cfg.correction.coef[i]=d;
	}

      xmms_cfg_free(cfile);
    }

#ifdef debug
  printf("READ CONFIG\n");
  sonyrm_dumpconfig(&sonyrm_cfg);
#endif
}

/** writes the global configuration structure to XMMS config file */
void sonyrm_save_config(void)
{
  int i;

  ConfigFile *cfile = xmms_cfg_open_default_file();
  if (!cfile)
    cfile = xmms_cfg_new();

  xmms_cfg_write_string(cfile, "sonystick", "device", sonyrm_cfg.device);

  for(i=0; i<SONY_NUM; i++)
    xmms_cfg_write_int(cfile, "sonystick", sonyrm_configstr[i], sonyrm_cfg.command[i]);

  xmms_cfg_write_int(cfile, "sonystick", "bitmask", sonyrm_cfg.bitmask);

  xmms_cfg_write_int(cfile, "sonystick", "type", sonyrm_cfg.correction.type);
  xmms_cfg_write_int(cfile, "sonystick", "prec", sonyrm_cfg.correction.prec);
  for(i=0; i<8; i++)
    xmms_cfg_write_int(cfile, "sonystick", sonyrm_configstr[SONY_NUM+i], sonyrm_cfg.correction.coef[i]);

  xmms_cfg_write_default_file(cfile);
  xmms_cfg_free(cfile);

#ifdef debug
  printf("WRITE CONFIG\n");
  sonyrm_dumpconfig(&sonyrm_cfg);
#endif
}

/** initialize the plugin and start it */
void sonyrm_init()
{
  static int fd;

  sonyrm_read_config();
  if((fd=sonyrm_open(&sonyrm_cfg)) < 0)
    return;

  keep_going = TRUE;
  pthread_create(&sonyrm_thread, NULL, sonyrm_routine, &fd);
}

/** stop plugin */
void sonyrm_cleanup()
{
  keep_going = FALSE;
  pthread_join(sonyrm_thread, NULL);
}

/** the thread function, which processes input from the joystick port 
    @param p_fd pointer to filehandle of joystick port
*/
static void *sonyrm_routine(void *p_fd)
{
  /* gint output_time, playlist_pos, playlist_time, playlist_length; */

  gboolean button_pressed=FALSE;
  gboolean volume_down=FALSE;
  gboolean volume_up=FALSE;
  gboolean aot=FALSE;

  gint i;

  int fd = *((int *)p_fd);	

  if(fd<0)
    return p_fd;

  while (keep_going)
    {
      fd_set set;
      struct js_event js;
      struct timeval tv;

      FD_ZERO(&set);
      FD_SET(fd, &set);

      tv.tv_sec = 0;
      tv.tv_usec = 25000;

      if (select(fd+1, &set, NULL, NULL, &tv))
	{
	  if (FD_ISSET(fd, &set))
	    {
	      gint joyvalue=-1;
	      gint command=-1;

	      if (read(fd, &js, sizeof (struct js_event)) != sizeof (struct js_event))
		perror("\nSonystick Control: error reading");

	      switch (js.type)
		{
		case JS_EVENT_BUTTON: 
		  button_pressed=!button_pressed;
		  if(!button_pressed)
		    joyvalue=SONY_BUTTON; 
		  break;
		case JS_EVENT_AXIS: joyvalue=js.value&sonyrm_cfg.bitmask; break;
		}

	      for(i=0; i<SONY_NUM; i++)
		if(sonyrm_cfg.command[i]==joyvalue)
		  {
		    command=i;
		    break;
		  }

#ifdef debug
	      fprintf(stderr, "joyvalue: %i -> command: %i\n", joyvalue, command);
#endif

	      switch(command)
		{
		case SONY_EJECT:
		  xmms_remote_eject(sonyrm_gp.xmms_session);
		  break;
		case SONY_PLAY:
		  if (xmms_remote_is_playing(sonyrm_gp.xmms_session))
		    xmms_remote_pause(sonyrm_gp.xmms_session);
		  else
		    xmms_remote_play(sonyrm_gp.xmms_session);
		  break;
		case SONY_STOP:
		  xmms_remote_stop(sonyrm_gp.xmms_session);
		  break;
		case SONY_NEXT:
		  xmms_remote_playlist_next(sonyrm_gp.xmms_session);
		  break;
		case SONY_PREV:
		  xmms_remote_playlist_prev(sonyrm_gp.xmms_session);
		  break;
		case SONY_VOLUP:
		  volume_up=!volume_up; break;
		case SONY_VOLDOWN:
		  volume_down=!volume_down; break;
		case SONY_REPEAT:
		  xmms_remote_toggle_repeat(sonyrm_gp.xmms_session);
		  break;
		case SONY_SHUFFLE:
		  xmms_remote_toggle_shuffle(sonyrm_gp.xmms_session);
		  break;
		case SONY_AOT:
		  aot=!aot;
		  xmms_remote_toggle_aot(sonyrm_gp.xmms_session, aot);
		  break;
		case SONY_EQWIN:
		  xmms_remote_eq_win_toggle(sonyrm_gp.xmms_session, !xmms_remote_is_eq_win(sonyrm_gp.xmms_session));
		  break;
		case SONY_MAINWIN:
		  xmms_remote_main_win_toggle(sonyrm_gp.xmms_session, !xmms_remote_is_main_win(sonyrm_gp.xmms_session));
		  break;
		case SONY_PLWIN:
		  xmms_remote_pl_win_toggle(sonyrm_gp.xmms_session, !xmms_remote_is_pl_win(sonyrm_gp.xmms_session));
		  break;
		case SONY_PREFWIN:
		  xmms_remote_show_prefs_box(sonyrm_gp.xmms_session);
		  break;

		case -1:
		default: 
		  volume_down=0;      
		  volume_up=0;	      
		  break;
		}

	    }
	  /*
	    case SONY_FWD5:
	    playlist_pos = xmms_remote_get_playlist_pos(sonyrm_gp.xmms_session);
	    playlist_length = xmms_remote_get_playlist_length(sonyrm_gp.xmms_session);
	    if (playlist_length - playlist_pos < 5)
	    playlist_pos = playlist_length - 5;
	    xmms_remote_set_playlist_pos(sonyrm_gp.xmms_session, playlist_pos + 5);
	    break;
	    case SONY_BWD5:
	    playlist_pos = xmms_remote_get_playlist_pos(sonyrm_gp.xmms_session);
	    if (playlist_pos < 5)
	    playlist_pos = 5;
	    xmms_remote_set_playlist_pos(sonyrm_gp.xmms_session, playlist_pos - 5);
	    break;
	    case SONY_FWD:
	    output_time = xmms_remote_get_output_time(sonyrm_gp.xmms_session);
	    playlist_pos = xmms_remote_get_playlist_pos(sonyrm_gp.xmms_session);
	    playlist_time = xmms_remote_get_playlist_time(sonyrm_gp.xmms_session, playlist_pos);
	    if (playlist_time - output_time < 5000)
	    output_time = playlist_time - 5000;
	    xmms_remote_jump_to_time(sonyrm_gp.xmms_session, output_time + 5000);
	    break;
	    case SONY_RWD:
	    output_time = xmms_remote_get_output_time(sonyrm_gp.xmms_session);
	    if (output_time < 5000)
	    output_time = 5000;
	    xmms_remote_jump_to_time(sonyrm_gp.xmms_session, output_time - 5000);
	    break;
	  */
	}

      if(volume_up)
	xmms_remote_set_main_volume(sonyrm_gp.xmms_session, xmms_remote_get_main_volume(sonyrm_gp.xmms_session)+1);
      if(volume_down)
	xmms_remote_set_main_volume(sonyrm_gp.xmms_session, xmms_remote_get_main_volume(sonyrm_gp.xmms_session)-1);

    }

  if(close(fd) < 0)
    perror("sonyrm_thread(): could not close joystick port");

  pthread_exit(NULL);
}

/** copies the sonyrm configuration structure. No memory is allocated or freed.
    @param dest destination struct
    @param src source struct
*/
void sonyrm_copy(sonyrm_config *dest, sonyrm_config *src)
{
  int i;

  g_free(dest->device);
  dest->device=g_strdup(src->device);

  for(i=0; i<SONY_NUM; i++)
    dest->command[i]=src->command[i];

  dest->bitmask=src->bitmask;

  dest->correction.prec=src->correction.prec;
  dest->correction.type=src->correction.type;
  for(i=0; i<8; i++)
    dest->correction.coef[i]=src->correction.coef[i];
}

/** opens the joystick port. The device is read from the configuration cfg.
    The port is calibrated with values read from cfg
    @param cfg a pointer to a sonyrm configuration structure
    @param nonblock true: open device on nonblock mode
    @return -1 on error. a filehandle on success
*/
int sonyrm_open(sonyrm_config *cfg)
{
  int fd=-1;
  char val;

  if ((fd = open(cfg->device, O_RDONLY)) < 0)
    {
      perror("Sonystick Control");
      return -1;
    }

  if (ioctl(fd, JSIOCGAXES, &val)) 
    {
      perror("error getting axes");
      if(close(fd) < 0)
	perror("error while closing");
      return -1;
    }
  
  if(val!=1)
    {
      fprintf(stderr, "joystick has %i axes\n", val);
      if(close(fd) < 0)
	perror("error while closing");
      return -1;
    }

  if (ioctl(fd, JSIOCGBUTTONS, &val)) 
    {
      perror("error getting buttons");
      if(close(fd) < 0)
	perror("error while closing");
      return -1;
    }

  if(val!=1)
    {
      fprintf(stderr, "joystick has %i buttons\n", val);
      if(close(fd) < 0)
	perror("error while closing");
      return -1;
    }

  /*
  if(cfg->correction.type!=SONY_IGNORE_CALIBRATION)
    {
      if (ioctl(fd, JSIOCSCORR, &cfg->correction)) 
	{
	  perror("error setting correction");
	  if(close(fd) < 0)
	    perror("error while closing");
	  return -1;
	}
    }
  */

  return fd;
}

/** prints the config structure
    @param cfg config structure
*/
void sonyrm_dumpconfig(sonyrm_config *cfg)
{
  int i;
  printf("Device:\t%s\n", cfg->device);
  for(i=0; i<SONY_NUM; i++)
    printf("%s:\t%i\n", sonyrm_configstr[i], cfg->command[i]);
  printf("bitmask:\t%#x\n", cfg->bitmask);
  printf("type:\t%i\n", cfg->correction.type);
  printf("prec:\t%i\n", cfg->correction.prec);
  for(i=0; i<8; i++)
    printf("coef%i:\t%i\n", i, cfg->correction.coef[i]);
}
